package eu.eudat.file.transformer.service.wordfiletransformer.word;

import eu.eudat.commonmodels.enums.FieldType;
import eu.eudat.commonmodels.models.description.DescriptionModel;
import eu.eudat.commonmodels.models.description.PropertyDefinitionFieldSetItemModel;
import eu.eudat.commonmodels.models.description.PropertyDefinitionFieldSetModel;
import eu.eudat.commonmodels.models.description.PropertyDefinitionModel;
import eu.eudat.commonmodels.models.descriptiotemplate.*;
import eu.eudat.commonmodels.models.descriptiotemplate.fielddata.*;
import eu.eudat.commonmodels.models.dmp.DmpModel;
import eu.eudat.commonmodels.models.dmpreference.DmpReferenceModel;
import eu.eudat.commonmodels.models.reference.ReferenceFieldModel;
import eu.eudat.commonmodels.models.reference.ReferenceModel;
import eu.eudat.file.transformer.service.storage.FileStorageServiceProperties;
import eu.eudat.file.transformer.service.wordfiletransformer.WordFileTransformerServiceProperties;
import eu.eudat.file.transformer.model.PidLink;
import eu.eudat.file.transformer.model.interfaces.ApplierWithValue;
import eu.eudat.file.transformer.service.pid.PidService;
import eu.eudat.file.transformer.model.enums.ParagraphStyle;
import org.apache.poi.openxml4j.exceptions.InvalidFormatException;
import org.apache.poi.util.Units;
import org.apache.poi.xwpf.usermodel.*;
import org.apache.xmlbeans.XmlCursor;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.select.NodeTraversor;
import org.openxmlformats.schemas.wordprocessingml.x2006.main.CTAbstractNum;
import org.openxmlformats.schemas.wordprocessingml.x2006.main.CTLvl;
import org.openxmlformats.schemas.wordprocessingml.x2006.main.STNumberFormat;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.stream.ImageInputStream;
import javax.management.InvalidApplicationException;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.*;

import static org.apache.poi.xwpf.usermodel.Document.*;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class WordBuilderImpl implements WordBuilder {
    private static final Logger logger = LoggerFactory.getLogger(WordBuilderImpl.class);
    private static final Map<String, Integer> IMAGE_TYPE_MAP = Map.of(
            "image/jpeg", PICTURE_TYPE_JPEG,
            "image/png", PICTURE_TYPE_PNG,
            "image/gif", PICTURE_TYPE_GIF,
            "image/tiff", PICTURE_TYPE_TIFF,
            "image/bmp", PICTURE_TYPE_BMP,
            "image/wmf", PICTURE_TYPE_WMF
    );
    private BigInteger numId;
    private Integer indent;
    private Integer imageCount;
    private final CTAbstractNum cTAbstractNum;
    private final FileStorageServiceProperties fileStorageServiceProperties;
    private final WordFileTransformerServiceProperties wordFileTransformerServiceProperties;
    private final PidService pidService;
    private final Map<ParagraphStyle, ApplierWithValue<XWPFDocument, Object, XWPFParagraph>> options = new HashMap<>();
    private final Map<ParagraphStyle, ApplierWithValue<XWPFTableCell, Object, XWPFParagraph>> optionsInTable = new HashMap<>();

    public WordBuilderImpl(FileStorageServiceProperties fileStorageServiceProperties, WordFileTransformerServiceProperties wordFileTransformerServiceProperties, PidService pidService) {
        this.fileStorageServiceProperties = fileStorageServiceProperties;
	    this.wordFileTransformerServiceProperties = wordFileTransformerServiceProperties;
	    this.pidService = pidService;
	    this.cTAbstractNum = CTAbstractNum.Factory.newInstance();
        this.cTAbstractNum.setAbstractNumId(BigInteger.valueOf(1));
        this.indent = 0;
        this.imageCount = 0;
        this.buildOptions();
        this.buildOptionsInTable();
    }

    private void buildOptionsInTable() {
        this.optionsInTable.put(ParagraphStyle.TEXT, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.addParagraph();
            XWPFRun run = paragraph.createRun();
            if (item != null)
                run.setText("" + item);
            run.setFontSize(11);
            return paragraph;
        });
        this.optionsInTable.put(ParagraphStyle.HTML, (mainDocumentPart, item) -> {
            Document htmlDoc = Jsoup.parse(((String)item).replaceAll("\n", "<br>"));
            HtmlToWorldBuilder htmlToWorldBuilder = HtmlToWorldBuilder.convertInTable(mainDocumentPart, htmlDoc, 0);
            return htmlToWorldBuilder.getParagraph();
        });
        this.optionsInTable.put(ParagraphStyle.TITLE, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.addParagraph();
            paragraph.setStyle("Title");
            paragraph.setAlignment(ParagraphAlignment.CENTER);
            XWPFRun run = paragraph.createRun();
            run.setText((String)item);
            run.setBold(true);
            run.setFontSize(14);
            return paragraph;
        });
        this.optionsInTable.put(ParagraphStyle.IMAGE, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.addParagraph();
            XWPFRun run = paragraph.createRun();
            if (item != null)
                run.setText(((Map<String, String>)item).get("name"));
            run.setFontSize(11);
            run.setItalic(true);
            return paragraph;
        });
    }

    private void buildOptions() {
        this.options.put(ParagraphStyle.TEXT, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.createParagraph();
            XWPFRun run = paragraph.createRun();
            if (item != null)
                run.setText("" + item);
            run.setFontSize(11);
            return paragraph;
        });
        this.options.put(ParagraphStyle.HTML, (mainDocumentPart, item) -> {
            Document htmlDoc = Jsoup.parse(((String)item).replaceAll("\n", "<br>"));
            HtmlToWorldBuilder htmlToWorldBuilder = HtmlToWorldBuilder.convert(mainDocumentPart, htmlDoc, this.indent);
            return htmlToWorldBuilder.getParagraph();
        });
        this.options.put(ParagraphStyle.TITLE, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.createParagraph();
            paragraph.setStyle("Title");
            paragraph.setAlignment(ParagraphAlignment.CENTER);
            XWPFRun run = paragraph.createRun();
            run.setText((String)item);
            run.setBold(true);
            run.setFontSize(14);
            return paragraph;
        });
        this.options.put(ParagraphStyle.HEADER1, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.createParagraph();
            paragraph.setStyle("Heading1");
            XWPFRun run = paragraph.createRun();
            run.setText((String)item);
            return paragraph;
        });
        this.options.put(ParagraphStyle.HEADER2, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.createParagraph();
            paragraph.setStyle("Heading2");
            XWPFRun run = paragraph.createRun();
            run.setText("" + item);
            return paragraph;
        });
        this.options.put(ParagraphStyle.HEADER3, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.createParagraph();
            paragraph.setStyle("Heading3");
            XWPFRun run = paragraph.createRun();
            run.setText("" + item);
            return paragraph;
        });
        this.options.put(ParagraphStyle.HEADER4, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.createParagraph();
            paragraph.setStyle("Heading4");
            XWPFRun run = paragraph.createRun();
            run.setText((String)item);
            return paragraph;
        });
        this.options.put(ParagraphStyle.HEADER5, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.createParagraph();
            paragraph.setStyle("Heading5");
            XWPFRun run = paragraph.createRun();
            run.setText("" + item);
            return paragraph;
        });
        this.options.put(ParagraphStyle.HEADER6, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.createParagraph();
            paragraph.setStyle("Heading6");
            XWPFRun run = paragraph.createRun();
            run.setText("" + item);
            return paragraph;
        });
        this.options.put(ParagraphStyle.FOOTER, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.createParagraph();
            XWPFRun run = paragraph.createRun();
            run.setText((String)item);
            return paragraph;
        });
        this.options.put(ParagraphStyle.COMMENT, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.createParagraph();
            XWPFRun run = paragraph.createRun();
            run.setText("" + item);
            run.setItalic(true);
            return paragraph;
        });
        this.options.put(ParagraphStyle.IMAGE, (mainDocumentPart, item) -> {
            XWPFParagraph paragraph = mainDocumentPart.createParagraph();
            paragraph.setPageBreak(true);
            paragraph.setSpacingAfter(0);
            paragraph.setAlignment(ParagraphAlignment.CENTER); //GK: Center the image if it is too small
            XWPFRun run = paragraph.createRun();
            String imageId = ((Map<String, String>)item).get("id");
            String fileName = ((Map<String, String>)item).get("name");
            String fileType = ((Map<String, String>)item).get("type");
            int format;
            format = IMAGE_TYPE_MAP.getOrDefault(fileType, 0);
            try {
                FileInputStream image = new FileInputStream(fileStorageServiceProperties.getTemp() + imageId);
                ImageInputStream iis = ImageIO.createImageInputStream(new File(fileStorageServiceProperties.getTemp() + imageId));
                Iterator<ImageReader> readers = ImageIO.getImageReaders(iis);
                if (readers.hasNext()) {
                    ImageReader reader = readers.next();
                    reader.setInput(iis);

                    int initialImageWidth = reader.getWidth(0);
                    int initialImageHeight = reader.getHeight(0);

                    float ratio = initialImageHeight / (float)initialImageWidth;

                    int marginLeftInDXA = (int) mainDocumentPart.getDocument().getBody().getSectPr().getPgMar().getLeft();
                    int marginRightInDXA = (int) mainDocumentPart.getDocument().getBody().getSectPr().getPgMar().getRight();
                    int pageWidthInDXA = (int) mainDocumentPart.getDocument().getBody().getSectPr().getPgSz().getW();
                    int pageWidth = Math.round((pageWidthInDXA - marginLeftInDXA - marginRightInDXA) / (float)20); // /20 converts dxa to points

                    int imageWidth = Math.round(initialImageWidth*(float)0.75);    // *0.75 converts pixels to points
                    int width = Math.min(imageWidth, pageWidth);

                    int marginTopInDXA = (int) mainDocumentPart.getDocument().getBody().getSectPr().getPgMar().getTop();
                    int marginBottomInDXA = (int) mainDocumentPart.getDocument().getBody().getSectPr().getPgMar().getBottom();
                    int pageHeightInDXA = (int) mainDocumentPart.getDocument().getBody().getSectPr().getPgSz().getH();
                    int pageHeight = Math.round((pageHeightInDXA - marginTopInDXA - marginBottomInDXA) / (float)20);    // /20 converts dxa to points

                    int imageHeight = Math.round(initialImageHeight * ((float)0.75));  // *0.75 converts pixels to points

                    int height = Math.round(width*ratio);
                    if(height > pageHeight) {
                        // height calculated with ratio is too large. Image may have Portrait (vertical) orientation. Recalculate image dimensions.
                        height = Math.min(imageHeight, pageHeight);
                        width = Math.round(height/ratio);
                    }

                    run.addPicture(image, format, fileName, Units.toEMU(width), Units.toEMU(height));
                    paragraph.setPageBreak(false);
                    imageCount++;
                    XWPFParagraph captionParagraph = mainDocumentPart.createParagraph();
                    captionParagraph.setAlignment(ParagraphAlignment.CENTER);
                    captionParagraph.setSpacingBefore(0);
                    captionParagraph.setStyle("Caption");
                    XWPFRun captionRun = captionParagraph.createRun();
                    captionRun.setText("Image " + imageCount);

                }
            } catch (IOException | InvalidFormatException e){
                logger.error(e.getMessage(), e);
            }
            return paragraph;
        });
    }

    @Override
    public void build(XWPFDocument document, DescriptionTemplateModel descriptionTemplate, PropertyDefinitionModel propertyDefinitionModel) {
        createPages(descriptionTemplate.getDefinition().getPages(), propertyDefinitionModel, document);
    }

    private void createPages(List<PageModel> datasetProfilePages,  PropertyDefinitionModel propertyDefinitionModel, XWPFDocument mainDocumentPart) {
        datasetProfilePages.stream().filter(item -> item.getSections() != null).forEach(item -> {
            try {
                createSections(item.getSections(), propertyDefinitionModel, mainDocumentPart, 0, false, item.getOrdinal() + 1, null);
            } catch (Exception e) {
                logger.error(e.getMessage(), e);
            }
        });
    }

    private void createSections(List<SectionModel> sections,  PropertyDefinitionModel propertyDefinitionModel, XWPFDocument mainDocumentPart, Integer indent, Boolean createListing, Integer page, String sectionString) {
        if (createListing) this.addListing(indent, false, true);
        boolean hasValue = false;

        for (SectionModel section: sections) {
            int paragraphPos = -1;
            String tempSectionString = sectionString != null ? sectionString + "." + (section.getOrdinal() + 1) : "" + (section.getOrdinal() + 1);
            if (!createListing) {
                XWPFParagraph paragraph = addParagraphContent(page + "." + tempSectionString + " " + section.getTitle(), mainDocumentPart, ParagraphStyle.HEADER5, numId, indent);
                paragraphPos = mainDocumentPart.getPosOfParagraph(paragraph);
            }
            if (section.getSections() != null) {
                createSections(section.getSections(), propertyDefinitionModel, mainDocumentPart, indent + 1, createListing, page, tempSectionString);
            }
            if (section.getFieldSets() != null) {
                hasValue = createFieldSetFields(section.getFieldSets(), propertyDefinitionModel, mainDocumentPart, indent + 1, createListing, page, tempSectionString);
            }

            if (!hasValue && paragraphPos > -1) {
                mainDocumentPart.removeBodyElement(paragraphPos);
            }
        }
    }

    private Boolean createFieldSetFields(List<FieldSetModel> fieldSetModels, PropertyDefinitionModel propertyDefinitionModel, XWPFDocument mainDocumentPart, Integer indent, Boolean createListing, Integer page, String section) {
        if (createListing) this.addListing(indent, true, true);
        boolean hasValue = false;
        boolean returnedValue = false;

        for (FieldSetModel fieldSetModel : fieldSetModels) {
            PropertyDefinitionFieldSetModel propertyDefinitionFieldSetModel = propertyDefinitionModel.getFieldSets().getOrDefault(fieldSetModel.getId(), null);
            List<PropertyDefinitionFieldSetItemModel> propertyDefinitionFieldSetItemModels = propertyDefinitionFieldSetModel != null && propertyDefinitionFieldSetModel.getItems() != null ? propertyDefinitionFieldSetModel.getItems() : new ArrayList<>();
            propertyDefinitionFieldSetItemModels = propertyDefinitionFieldSetItemModels.stream().sorted(Comparator.comparingInt(PropertyDefinitionFieldSetItemModel::getOrdinal)).toList();

            char c = 'a';
            int multiplicityItems = 0;
            boolean hasMultiplicityItems = false;
            int paragraphPos = -1;

            if (fieldSetModel.getTitle() != null && !fieldSetModel.getTitle().isEmpty() && !createListing) {
                XWPFParagraph paragraph = addParagraphContent(page + "." + section + "." + (fieldSetModel.getOrdinal() + 1) + " " + fieldSetModel.getTitle(), mainDocumentPart, ParagraphStyle.HEADER6, numId, indent);
                paragraphPos = mainDocumentPart.getPosOfParagraph(paragraph);
            }

            if (!propertyDefinitionFieldSetItemModels.isEmpty()) {
                if (fieldSetModel.getMultiplicity() != null) {
                    if (fieldSetModel.getMultiplicity().getTableView()) {
                        XWPFTable tbl = mainDocumentPart.createTable();
                        int numOfRows = 0;
                        
                        tbl.setTableAlignment(TableRowAlign.CENTER);
                        mainDocumentPart.createParagraph();
                        createHeadersInTable(fieldSetModel.getFields(), tbl);
                        numOfRows = tbl.getRows().size();

                        for (PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel : propertyDefinitionFieldSetItemModels) {
                            XWPFTableRow row = tbl.createRow();
                            hasValue = createFieldsInTable(fieldSetModel, propertyDefinitionFieldSetItemModel, row, indent, createListing, false, numOfRows);
                            numOfRows++;
                            if (hasValue) returnedValue = true;
                        }
                    } else {
                        for (PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel : propertyDefinitionFieldSetItemModels) {
                            int paragraphPosInner = -1;
                            if (!fieldSetModel.getMultiplicity().getTableView() && !createListing) {
                                XWPFParagraph paragraphInner = addParagraphContent(c + ". ", mainDocumentPart, ParagraphStyle.TEXT, numId, indent);
                                c++;
                                paragraphPosInner = mainDocumentPart.getPosOfParagraph(paragraphInner);
                                hasMultiplicityItems = true;
                                multiplicityItems++;
                            }
                            hasValue = createFields(fieldSetModel, propertyDefinitionFieldSetItemModel, mainDocumentPart, indent, createListing, hasMultiplicityItems);
                            if (hasValue) {
                                returnedValue = true;
                            } else if (paragraphPosInner > -1) {
                                mainDocumentPart.removeBodyElement(paragraphPosInner);
                                c--;
                                multiplicityItems--;
                            }
                        }
                    }
                    if (multiplicityItems == 1) {
                        String text = mainDocumentPart.getLastParagraph().getRuns().getFirst().getText(0);
                        if (text.equals("a. ")) {
                            mainDocumentPart.getLastParagraph().removeRun(0);
                        }
                    }
                } else {
                    hasValue = createFields(fieldSetModel, propertyDefinitionFieldSetItemModels.getFirst(), mainDocumentPart, indent, createListing, false);
                }
                if (hasValue && propertyDefinitionFieldSetItemModels.getFirst().getComment() != null && !propertyDefinitionFieldSetItemModels.getFirst().getComment().isEmpty() && !createListing) {
                    addParagraphContent("<i>Comment:</i>\n"+ propertyDefinitionFieldSetItemModels.getFirst().getComment(), mainDocumentPart, ParagraphStyle.HTML, numId, indent);
                }
            }
           
            if (!hasValue && paragraphPos > -1) {
                mainDocumentPart.removeBodyElement(paragraphPos);
            }
        }

        return returnedValue;
    }

    private void createHeadersInTable(List<FieldModel> fields, XWPFTable table) {
        boolean atLeastOneHeader = false;
        List<FieldModel> tempFields = fields.stream().sorted(Comparator.comparingInt(FieldModel::getOrdinal)).toList();
        int index = 0;
        XWPFTableRow row = table.getRow(0);
        for (FieldModel field: tempFields) {
            if (field.getIncludeInExport()) {
                XWPFTableCell cell;
                if (index == 0) {
                    cell = row.getCell(0);
                } else {
                    cell = row.createCell();
                }
                cell.setVerticalAlignment(XWPFTableCell.XWPFVertAlign.valueOf("CENTER"));
                String label = field.getData().getLabel();
                if (label != null && !label.isBlank()) {
                    XWPFParagraph paragraph = cell.getParagraphs().getFirst();
                    paragraph.setIndentationFirstLine(50);
                    XWPFRun run = paragraph.createRun();
                    run.setText(label);
                    run.setBold(true);
                    run.setFontSize(12);
                    paragraph.setAlignment(ParagraphAlignment.CENTER);
                    paragraph.setSpacingBefore(100);

                    atLeastOneHeader = true;
                }
            }
            index++;
        }

        if(!atLeastOneHeader) {
            table.removeRow(0);
        }
    }

    private Boolean createFieldsInTable(FieldSetModel fieldSetModel, PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel, XWPFTableRow mainDocumentPart, Integer indent, Boolean createListing, boolean hasMultiplicityItems, int numOfRows) {
        int numOfCells = 0;
        boolean hasValue = false;
        List<FieldModel> tempFields = fieldSetModel.getFields().stream().sorted(Comparator.comparingInt(FieldModel::getOrdinal)).toList();
        for (FieldModel field: tempFields) {
            if (field.getIncludeInExport()) {
                if (!createListing) {
                    try {
                        eu.eudat.commonmodels.models.description.FieldModel fieldValueModel = propertyDefinitionFieldSetItemModel.getFields().getOrDefault(field.getId(), null);
                        if(field.getData().getFieldType().equals(FieldType.UPLOAD)){
                            boolean isImage = false;
                            for(UploadDataModel.UploadOptionModel type: ((UploadDataModel)field.getData()).getTypes()){
                                String fileFormat = type.getValue();
                                if(IMAGE_TYPE_MAP.containsKey(fileFormat)){
                                    isImage = true;
                                    break;
                                }
                            }
                            if(isImage){
                                if (fieldValueModel != null && fieldValueModel.getTextValue() != null && !fieldValueModel.getTextValue().isEmpty()) {
                                    XWPFParagraph paragraph = addCellContent(fieldValueModel.getTextValue(), mainDocumentPart, ParagraphStyle.IMAGE, numId, 0, numOfRows, numOfCells, 0); //TODO
                                    if (paragraph != null) {
                                        hasValue = true;
                                    }
                                    if(hasMultiplicityItems){
                                        hasMultiplicityItems = false;
                                    }
                                }
                            }
                        }
                        else if (fieldValueModel != null) {
                            this.indent = indent;
                            String format = this.formatter(field, fieldValueModel);
                            Boolean hasMultiAutoComplete = false;
                            boolean isResearcher = false;
                            if (field.getData() instanceof LabelAndMultiplicityDataModel){
                                hasMultiAutoComplete = ((LabelAndMultiplicityDataModel)field.getData()).getMultipleSelect();
                            } 
                            if (field.getData() instanceof SelectDataModel){
                                hasMultiAutoComplete = ((SelectDataModel)field.getData()).getMultipleSelect();
                            }
                            if (field.getData() instanceof ReferenceTypeDataModel){
                                hasMultiAutoComplete = ((ReferenceTypeDataModel)field.getData()).getMultipleSelect();
                                isResearcher = ((ReferenceTypeDataModel)field.getData()).getReferenceType().getCode().equals(this.wordFileTransformerServiceProperties.getResearcherReferenceCode());
                            }
                            
                            if(format != null && !format.isEmpty()){
                                boolean isMultiAutoComplete = hasMultiAutoComplete != null && hasMultiAutoComplete;
                                boolean arrayStringFormat = format.charAt(0) == '[';
                                if(arrayStringFormat || isMultiAutoComplete){
                                    List<String> values = (arrayStringFormat) ? Arrays.asList(format.substring(1, format.length() - 1).split(",[ ]*")) : Arrays.asList(format.split(",[ ]*"));
                                    if(values.size() > 1) {
                                        boolean orcidResearcher;
                                        int numOfValuesInCell = 0;
                                        for (String val : values) {
                                            orcidResearcher = false;
                                            String orcId = null;
                                            if(isResearcher && val.contains("orcid:")){
                                                orcId = val.substring(val.indexOf(':') + 1, val.indexOf(')'));
                                                val = val.substring(0, val.indexOf(':') + 1) + " ";
                                                orcidResearcher = true;
                                            }
                                            format = "• " + val;
                                            if(hasMultiplicityItems){
                                                XWPFParagraph paragraph = mainDocumentPart.getCell(mainDocumentPart.getTableCells().size()).addParagraph();
                                                paragraph.createRun().setText(format);
                                                if(orcidResearcher){
                                                    XWPFHyperlinkRun run = paragraph.createHyperlinkRun("https://orcid.org/" + orcId);
                                                    run.setText(orcId);
                                                    run.setUnderline(UnderlinePatterns.SINGLE);
                                                    run.setColor("0000FF");
                                                    paragraph.createRun().setText(")");
                                                }
                                                hasMultiplicityItems = false;
                                            }
                                            else{
                                                XWPFParagraph paragraph = addCellContent(format, mainDocumentPart, field.getData().getFieldType().equals(FieldType.RICH_TEXT_AREA) ? ParagraphStyle.HTML : ParagraphStyle.TEXT, numId, indent, numOfRows, numOfCells, numOfValuesInCell);
                                                numOfValuesInCell++;
                                                if(orcidResearcher){
                                                    XWPFHyperlinkRun run = paragraph.createHyperlinkRun("https://orcid.org/" + orcId);
                                                    run.setText(orcId);
                                                    run.setUnderline(UnderlinePatterns.SINGLE);
                                                    run.setColor("0000FF");
                                                    paragraph.createRun().setText(")");
                                                }
                                                if (paragraph != null) {
                                                    hasValue = true;
                                                }
                                            }
                                            format = null;
                                        }
                                    }
                                    else if(values.size() == 1){
                                        format = values.getFirst();
                                    }
                                }
                            }
                            if(hasMultiplicityItems && format != null){
                                XWPFParagraph paragraph = mainDocumentPart.getCell(mainDocumentPart.getTableCells().size()).addParagraph();
                                paragraph.createRun().setText(format);
                                hasMultiplicityItems = false;
                                hasValue = true;
                            }
                            else{
                                XWPFParagraph paragraph = addCellContent(format, mainDocumentPart, field.getData().getFieldType().equals(FieldType.RICH_TEXT_AREA) ? ParagraphStyle.HTML : ParagraphStyle.TEXT, numId, indent, numOfRows, numOfCells, 0);
                                if (paragraph != null) {
                                    hasValue = true;
                                }
                            }
                        }
                    } catch (InvalidApplicationException e) {
                        logger.error(e.getMessage(), e);
                    }
                }
                numOfCells++;
            }
        }
        return hasValue;
    }

    private void createHypeLink(XWPFDocument mainDocumentPart, String format, String pidType, String pid, boolean hasMultiplicityItems, boolean isMultiAutoComplete){
        PidLink pidLink = pidService.getPid(pidType);
        if (pidLink != null) {
            if (!hasMultiplicityItems) {
                XWPFParagraph paragraph = mainDocumentPart.createParagraph();
                paragraph.setIndentFromLeft(400 * indent);
                if (numId != null) {
                    paragraph.setNumID(numId);
                }
            }
            if (isMultiAutoComplete) {
                XWPFRun r = mainDocumentPart.getLastParagraph().createRun();
                r.setText("• ");
            }
            XWPFHyperlinkRun run = mainDocumentPart.getLastParagraph().createHyperlinkRun(pidLink.getLink().replace("{pid}", pid));
            run.setText(format);
            run.setUnderline(UnderlinePatterns.SINGLE);
            run.setColor("0000FF");
            run.setFontSize(11);
        }
        else {
            String newFormat = (isMultiAutoComplete) ? "• " + format : format;
            if (hasMultiplicityItems) {
                mainDocumentPart.getLastParagraph().createRun().setText(newFormat);
            }
            else {
                addParagraphContent(newFormat, mainDocumentPart, ParagraphStyle.TEXT, numId, indent);
            }
        }
    }

    private Boolean createFields(FieldSetModel fieldSetModel, PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel, XWPFDocument mainDocumentPart, Integer indent, Boolean createListing, boolean hasMultiplicityItems) {
        if (createListing) this.addListing(indent, false, false);
        boolean hasValue = false;
        List<FieldModel> tempFields = fieldSetModel.getFields().stream().sorted(Comparator.comparingInt(FieldModel::getOrdinal)).toList();
        for (FieldModel field: tempFields) {
            if (field.getIncludeInExport()) {
                if (!createListing) {
                    try {
                        eu.eudat.commonmodels.models.description.FieldModel fieldValueModel = propertyDefinitionFieldSetItemModel.getFields().getOrDefault(field.getId(), null);
                        if (field.getData() != null) {
                            if (field.getData().getFieldType().equals(FieldType.UPLOAD)) {
                                boolean isImage = false;
                                for (UploadDataModel.UploadOptionModel type : ((UploadDataModel) field.getData()).getTypes()) {
                                    String fileFormat = type.getValue();
                                    if (IMAGE_TYPE_MAP.containsKey(fileFormat)) {
                                        isImage = true;
                                        break;
                                    }
                                }
                                if (isImage) {
                                    if (fieldValueModel.getTextValue() != null && !fieldValueModel.getTextValue().isEmpty()) {
                                        XWPFParagraph paragraph = addParagraphContent(fieldValueModel.getTextValue(), mainDocumentPart, ParagraphStyle.IMAGE, numId, 0); //TODO
                                        if (paragraph != null) {
                                            hasValue = true;
                                        }
                                        if (hasMultiplicityItems) {
                                            hasMultiplicityItems = false;
                                        }
                                    }
                                }
                            } else if (fieldValueModel != null) {
                                this.indent = indent;
                                String format = this.formatter(field, fieldValueModel);
                                Boolean isMultiAutoComplete = false;
                                boolean isResearcher = false;
                                boolean isOrganization = false;
                                boolean isExternalDataset = false;
                                boolean isPublication = false;
                                if (field.getData() instanceof LabelAndMultiplicityDataModel) {
                                    isMultiAutoComplete = ((LabelAndMultiplicityDataModel) field.getData()).getMultipleSelect();
                                }
                                if (field.getData() instanceof SelectDataModel) {
                                    isMultiAutoComplete = ((SelectDataModel) field.getData()).getMultipleSelect();
                                }
                                if (field.getData() instanceof ReferenceTypeDataModel) {
                                    isMultiAutoComplete = ((ReferenceTypeDataModel) field.getData()).getMultipleSelect();
                                    isResearcher = ((ReferenceTypeDataModel) field.getData()).getReferenceType().getCode().equals(this.wordFileTransformerServiceProperties.getResearcherReferenceCode());
                                    isOrganization = ((ReferenceTypeDataModel) field.getData()).getReferenceType().getCode().equals(this.wordFileTransformerServiceProperties.getOrganizationReferenceCode());
                                    isExternalDataset = ((ReferenceTypeDataModel) field.getData()).getReferenceType().getCode().equals(this.wordFileTransformerServiceProperties.getDatasetReferenceCode());
                                    isPublication = ((ReferenceTypeDataModel) field.getData()).getReferenceType().getCode().equals(this.wordFileTransformerServiceProperties.getPublicationReferenceCode());
                                }

                                if (isOrganization || isExternalDataset || isPublication) {
                                    if(fieldValueModel.getReferences() != null){
                                        for (ReferenceModel referenceModel : fieldValueModel.getReferences())
                                            createHypeLink(mainDocumentPart, format, referenceModel.getDefinition().getFields().stream().filter(x-> x.getCode().equals("pidTypeField")).map(ReferenceFieldModel::getValue).findFirst().orElse(null), referenceModel.getReference(), hasMultiplicityItems, isMultiAutoComplete && fieldValueModel.getReferences().size() > 1);
                                        if (hasMultiplicityItems) hasMultiplicityItems = false;

                                        hasValue = true;
                                    }
                                    
                                } else {
                                    if (format != null && !format.isEmpty()) {
                                        boolean arrayStringFormat = format.charAt(0) == '[';
                                        if (arrayStringFormat || isMultiAutoComplete) {
                                            List<String> values = (arrayStringFormat) ? Arrays.asList(format.substring(1, format.length() - 1).split(",[ ]*")) : Arrays.asList(format.split(",[ ]*"));
                                            if (values.size() > 1) {
                                                boolean orcidResearcher;
                                                for (String val : values) {
                                                    orcidResearcher = false;
                                                    String orcId = null;
                                                    if (isResearcher && val.contains("orcid:")) {
                                                        orcId = val.substring(val.indexOf(':') + 1, val.indexOf(')'));
                                                        val = val.substring(0, val.indexOf(':') + 1) + " ";
                                                        orcidResearcher = true;
                                                    }
                                                    format = "• " + val;
                                                    if (hasMultiplicityItems) {
                                                        mainDocumentPart.getLastParagraph().createRun().setText(format);
                                                        if (orcidResearcher) {
                                                            XWPFHyperlinkRun run = mainDocumentPart.getLastParagraph().createHyperlinkRun("https://orcid.org/" + orcId);
                                                            run.setText(orcId);
                                                            run.setUnderline(UnderlinePatterns.SINGLE);
                                                            run.setColor("0000FF");
                                                            mainDocumentPart.getLastParagraph().createRun().setText(")");
                                                        }
                                                        hasMultiplicityItems = false;
                                                    } else {
                                                        XWPFParagraph paragraph = addParagraphContent(format, mainDocumentPart, field.getData().getFieldType().equals(FieldType.RICH_TEXT_AREA) ? ParagraphStyle.HTML : ParagraphStyle.TEXT, numId, indent);
                                                        if (orcidResearcher) {
                                                            XWPFHyperlinkRun run = paragraph.createHyperlinkRun("https://orcid.org/" + orcId);
                                                            run.setText(orcId);
                                                            run.setUnderline(UnderlinePatterns.SINGLE);
                                                            run.setColor("0000FF");
                                                            paragraph.createRun().setText(")");
                                                        }
                                                        if (paragraph != null) {
                                                            hasValue = true;
                                                        }
                                                    }
                                                    format = null;
                                                }
                                            } else if (values.size() == 1) {
                                                format = values.getFirst();
                                            }
                                        }
                                    }
                                    if (format != null) {
                                        if (hasMultiplicityItems) {
                                            mainDocumentPart.getLastParagraph().createRun().setText(format);
                                            hasMultiplicityItems = false;
                                            hasValue = true;
                                        } else {
                                            XWPFParagraph paragraph = addParagraphContent(format, mainDocumentPart, field.getData().getFieldType().equals(FieldType.RICH_TEXT_AREA) ? ParagraphStyle.HTML : ParagraphStyle.TEXT, numId, indent);
                                            if (paragraph != null) {
                                                hasValue = true;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    } catch (InvalidApplicationException e) {
	                    throw new RuntimeException(e);
                    }
                }
            }
        }
        return hasValue;
    }

    private XWPFParagraph addCellContent(Object content, XWPFTableRow mainDocumentPart, ParagraphStyle style, BigInteger numId, int indent, int numOfRows, int numOfCells, int numOfValuesInCell) {
        if (content != null) {
            if (content instanceof String && ((String)content).isEmpty()) {
                return null;
            }
            this.indent = indent;
            XWPFTableCell cell;
            if(numOfRows > 0 || numOfValuesInCell > 0) {
                cell = mainDocumentPart.getCell(numOfCells);
            } else {
                cell = mainDocumentPart.createCell();
            }
            cell.setVerticalAlignment(XWPFTableCell.XWPFVertAlign.valueOf("CENTER"));
            if(numOfValuesInCell == 0) {
                cell.removeParagraph(0);
            }

            XWPFParagraph paragraph = this.optionsInTable.get(style).apply(cell, content);
            if (paragraph != null) {
                paragraph.setAlignment(ParagraphAlignment.CENTER);
                paragraph.setSpacingBefore(100);
                if (numId != null) {
                    paragraph.setNumID(numId);
                }
                return paragraph;
            }
        }
        return null;
    }

    @Override
    public XWPFParagraph addParagraphContent(Object content, XWPFDocument mainDocumentPart, ParagraphStyle style, BigInteger numId, int indent) {
        if (content != null) {
            if (content instanceof String && ((String)content).isEmpty()) {
                return null;
            }
            this.indent = indent;
            XWPFParagraph paragraph = this.options.get(style).apply(mainDocumentPart, content);
            if (paragraph != null) {
                paragraph.setIndentFromLeft(400*indent);
                if (numId != null) {
                    paragraph.setNumID(numId);
                }
                return paragraph;
            }
        }
        return null;
    }

    private void addListing(int indent, boolean question, Boolean hasIndication) {
        CTLvl cTLvl = this.cTAbstractNum.addNewLvl();

        if (question) {
            cTLvl.addNewNumFmt().setVal(STNumberFormat.DECIMAL);
            cTLvl.setIlvl(BigInteger.valueOf(indent));
        } else {
            if (hasIndication) {
                cTLvl.addNewNumFmt().setVal(STNumberFormat.DECIMAL);
                cTLvl.setIlvl(BigInteger.valueOf(indent));
            } else {
                cTLvl.addNewNumFmt().setVal(STNumberFormat.NONE);
                cTLvl.setIlvl(BigInteger.valueOf(indent));
            }
        }
    }

    private String formatter(FieldModel field, eu.eudat.commonmodels.models.description.FieldModel fieldValueModel) throws InvalidApplicationException {
        if (fieldValueModel == null || field == null || field.getData() == null) {
            return null;
        }
        switch (field.getData().getFieldType()) {
            case REFERENCE_TYPES: {
                List<String> values = new ArrayList<>();
                if (fieldValueModel.getReferences() != null && !fieldValueModel.getReferences().isEmpty()) {
                    for (ReferenceModel referenceModel : fieldValueModel.getReferences()) {
                        if (referenceModel != null) {
                            if (referenceModel.getLabel() != null && !referenceModel.getLabel().isBlank()) {
                                values.add(referenceModel.getLabel());
                            }
                            if (referenceModel.getDescription() != null && !referenceModel.getDescription().isBlank()) {
                                values.add(referenceModel.getLabel());
                            }
                        }
                    }
                }

                return String.join(", ", values);
            }
            case TAGS:
            case CURRENCY: {
                List<String> values = new ArrayList<>();
                if (fieldValueModel.getTextListValue() != null && !fieldValueModel.getTextListValue().isEmpty()) {
                    values.addAll(fieldValueModel.getTextListValue());
                }
                return String.join(", ", values);
            }
            case SELECT: {
                List<String> values = new ArrayList<>();
                if (fieldValueModel.getTextListValue() != null && !fieldValueModel.getTextListValue().isEmpty()) {
                    SelectDataModel selectDataModel = (SelectDataModel)field.getData();
                    if (selectDataModel != null && selectDataModel.getOptions() != null && !selectDataModel.getOptions().isEmpty()){
                        for (SelectDataModel.OptionModel option : selectDataModel.getOptions()){
                            if (fieldValueModel.getTextListValue().contains(option.getValue()) || fieldValueModel.getTextListValue().contains(option.getLabel())) values.add(option.getLabel());
                        }
                    }
                }
                return String.join(", ", values);
            }
            case BOOLEAN_DECISION:
                if (fieldValueModel.getTextValue() != null && fieldValueModel.getTextValue().equals("true")) return "Yes";
                if (fieldValueModel.getTextValue() != null && fieldValueModel.getTextValue().equals("false")) return "No";
                return null;
            case RADIO_BOX:
                RadioBoxDataModel radioBoxDataModel = (RadioBoxDataModel) field.getData();
                if (fieldValueModel.getTextValue() != null && radioBoxDataModel != null && radioBoxDataModel.getOptions() != null) {
                    for (RadioBoxDataModel.RadioBoxOptionModel option : radioBoxDataModel.getOptions()){
                        if (option.getValue().equals(fieldValueModel.getTextValue()) || option.getLabel().equals(fieldValueModel.getTextValue())) return option.getLabel();
                    }
                }
                return null;
            case CHECK_BOX:
                LabelDataModel checkBoxData = (LabelDataModel) field.getData();
                if (fieldValueModel.getTextValue() != null || fieldValueModel.getTextValue().equals("false")) return null;
                return checkBoxData != null ? checkBoxData.getLabel() : null;
            case DATE_PICKER:{
                return fieldValueModel.getDateValue() != null ? DateTimeFormatter.ofPattern("yyyy-MM-dd").withZone(ZoneId.systemDefault()).format(fieldValueModel.getDateValue()) : "";
            }
            case FREE_TEXT:
            case TEXT_AREA:
            case RICH_TEXT_AREA:
                return fieldValueModel.getTextValue() != null ? fieldValueModel.getTextValue() : "";
            case DATASET_IDENTIFIER:
            case VALIDATION:
                if (fieldValueModel.getExternalIdentifier() != null) {
                    return "id: " + fieldValueModel.getExternalIdentifier().getIdentifier() + ", Type: " + fieldValueModel.getExternalIdentifier().getType();
                }
                return "";
            case UPLOAD:
            case INTERNAL_ENTRIES_DESCRIPTIONS:
            case INTERNAL_ENTRIES_DMPS: return null;
            default: throw new InvalidApplicationException("Invalid type " + field.getData().getFieldType()); 
        }
    }

    @Override
    public int findPosOfPoweredBy(XWPFDocument document) {
        if (document == null) throw new IllegalArgumentException("Document required");
        if (document.getParagraphs() == null) return -1;
        
        for (XWPFParagraph p : document.getParagraphs()) {
            List<XWPFRun> runs = p.getRuns();
            if (runs != null) {
                for (XWPFRun r : runs) {
                    String text = r.getText(0);
                    if (text != null) {
                        if (text.equals("Powered by")) {
                            return document.getPosOfParagraph(p) - 1;
                        }
                    }
                }
            }
        }
        return -1;
    }

    private List<ReferenceModel> getReferenceModelOfTypeCode(DmpModel dmp, String code){
        List<ReferenceModel> response = new ArrayList<>();
        if (dmp.getReferences() == null) return response;
        for (DmpReferenceModel dmpReferenceModel : dmp.getReferences()){
            if (dmpReferenceModel.getReference() != null && dmpReferenceModel.getReference().getType() != null && dmpReferenceModel.getReference().getType().getCode() != null  && dmpReferenceModel.getReference().getType().getCode().equals(code)){
                response.add(dmpReferenceModel.getReference());
            }
        }
        return response;
    }

    @Override
    public void fillFirstPage(DmpModel dmpEntity, DescriptionModel descriptionModel, XWPFDocument document, boolean isDescription) {
        if (dmpEntity == null) throw new IllegalArgumentException("DmpEntity required");
        if (document == null) throw new IllegalArgumentException("Document required");
        
        int parPos = 0;
        int descrParPos = -1;
        XWPFParagraph descrPar = null;
        List<ReferenceModel> grants = this.getReferenceModelOfTypeCode(dmpEntity, this.wordFileTransformerServiceProperties.getGrantReferenceCode());
        List<ReferenceModel> researchers = this.getReferenceModelOfTypeCode(dmpEntity, this.wordFileTransformerServiceProperties.getResearcherReferenceCode());
        List<ReferenceModel> organizations = this.getReferenceModelOfTypeCode(dmpEntity, this.wordFileTransformerServiceProperties.getOrganizationReferenceCode());
        List<ReferenceModel> funders = this.getReferenceModelOfTypeCode(dmpEntity, this.wordFileTransformerServiceProperties.getFunderReferenceCode());
        
        for(XWPFParagraph p: document.getParagraphs()){
            List<XWPFRun> runs = p.getRuns();
            if(runs != null){
                for(XWPFRun r : runs){
                    String text = r.getText(0);
                    if(text != null){
                        if(text.contains("{ARGOS.DMP.TITLE}")) {
                            text = text.replace("{ARGOS.DMP.TITLE}", dmpEntity.getLabel());
                            r.setText(text, 0);
                        } else if(text.contains("{ARGOS.DMP.VERSION}")) {
                            text = text.replace("{ARGOS.DMP.VERSION}", "Version " + dmpEntity.getVersion());
                            r.setText(text, 0);
                        } else if(descriptionModel != null && text.contains("{ARGOS.DATASET.TITLE}")) {
                            text = text.replace("{ARGOS.DATASET.TITLE}", descriptionModel.getLabel());
                            r.setText(text, 0);
                        } else if((text.contains("{ARGOS.DMP.DESCRIPTION}") && !isDescription) || (descriptionModel != null && text.contains("{ARGOS.DATASET.DESCRIPTION}") && isDescription)) {
                            descrParPos = parPos;
                            descrPar = p;
                            if(!isDescription) {
                                text = text.replace("{ARGOS.DMP.DESCRIPTION}", "");
                            } else {
                                text = text.replace("{ARGOS.DATASET.DESCRIPTION}", "");
                            }
                            r.setText(text, 0);
                        } else if(text.equals("{ARGOS.DMP.RESEARCHERS}")) {
                            StringBuilder researchersNames = new StringBuilder();
                            int i = 0;
                            for(ReferenceModel researcher : researchers){
                                i++;
                                researchersNames.append(researcher.getLabel()).append(i < researchers.size() ? ", " : "");
                            }
                            text = text.replace("{ARGOS.DMP.RESEARCHERS}", researchersNames.toString());
                            r.setText(text, 0);
                            r.setFontSize(17);
                        } else if(text.equals("{ARGOS.DMP.ORGANIZATIONS}")) {
                            StringBuilder organisationsNames = new StringBuilder();
                            int i = 0;
                            for(ReferenceModel organisation : organizations){
                                i++;
                                organisationsNames.append(organisation.getLabel()).append(i < organizations.size() ? ", " : "");
                            }
                            text = text.replace("{ARGOS.DMP.ORGANIZATIONS}", organisationsNames.toString());
                            r.setText(text, 0);
                            r.setFontSize(17);
                        }
                    }
                }
            }
            parPos++;
        }

        if((descrParPos != -1)  && (dmpEntity.getDescription() != null) && !isDescription) {
            XmlCursor cursor = descrPar.getCTP().newCursor();
            cursor.toNextSibling();
            Document htmlDoc = Jsoup.parse((dmpEntity.getDescription()).replaceAll("\n", "<br>"));
            HtmlToWorldBuilder htmlToWorldBuilder = new HtmlToWorldBuilder(descrPar, 0, cursor);
            NodeTraversor.traverse(htmlToWorldBuilder, htmlDoc);
        }
        if((descrParPos != -1) && (descriptionModel != null) && (descriptionModel.getDescription() != null) && isDescription) {
            XmlCursor cursor = descrPar.getCTP().newCursor();
            cursor.toNextSibling();
            Document htmlDoc = Jsoup.parse(( descriptionModel.getDescription()).replaceAll("\n", "<br>"));
            HtmlToWorldBuilder htmlToWorldBuilder = new HtmlToWorldBuilder(descrPar, 0, cursor);
            NodeTraversor.traverse(htmlToWorldBuilder, htmlDoc);
        }


        XWPFTable tbl = document.getTables().getFirst();
        Iterator<XWPFTableRow> it = tbl.getRows().iterator();
        it.next(); // skip first row
        if(it.hasNext() && !funders.isEmpty()){
            XWPFParagraph p = it.next().getCell(0).getParagraphs().getFirst();
            XWPFRun run = p.createRun();
            run.setText(funders.getFirst().getLabel());
            run.setFontSize(17);
            p.setAlignment(ParagraphAlignment.CENTER);
        }
        it = tbl.getRows().iterator();
        it.next();

        if(it.hasNext() && !grants.isEmpty()){
            XWPFParagraph p = it.next().getCell(1).getParagraphs().getFirst();
            XWPFRun run = p.createRun();
            String text = grants.getFirst().getLabel();
            String reference = grants.getFirst().getReference();
            if(reference != null) {
                String[] parts = reference.split("::");
                text += parts.length > 1 ? "/ No "+parts[parts.length - 1] : "";
            }
            run.setText(text);
            run.setFontSize(17);
            p.setAlignment(ParagraphAlignment.CENTER);
        }
    }

    @Override
    public void fillFooter(DmpModel dmpEntity, DescriptionModel descriptionModel, XWPFDocument document) {
        if (dmpEntity == null) throw new IllegalArgumentException("DmpEntity required");
        
        List<ReferenceModel> licences = this.getReferenceModelOfTypeCode(dmpEntity, this.wordFileTransformerServiceProperties.getLicenceReferenceCode());
        document.getFooterList().forEach(xwpfFooter -> {
            List<XWPFRun> runs = xwpfFooter.getParagraphs().getFirst().getRuns();
            if(runs != null){
                for(XWPFRun r : runs){
                    String text = r.getText(0);
                    if(text != null){
                        if(text.contains("{ARGOS.DMP.TITLE}")){
                            text = text.replace("{ARGOS.DMP.TITLE}", dmpEntity.getLabel());
                            r.setText(text, 0);
                        }
                        if(text.contains("{ARGOS.DATASET.TITLE}") && descriptionModel != null){
                            text = text.replace("{ARGOS.DATASET.TITLE}", descriptionModel.getLabel());
                            r.setText(text, 0);
                        }
                        if(text.contains("{ARGOS.DMP.LICENSE}")){
                            if (!licences.isEmpty()){
                                text = text.replace("{ARGOS.DMP.LICENSE}", licences.getFirst().getReference());
                            }
                            else {
                                text = text.replace("{ARGOS.DMP.LICENSE}", "License: -");
                            }
                            r.setText(text, 0);
                        }
                        if(text.contains("{ARGOS.DMP.DOI}")){
                            if(dmpEntity.getEntityDois() != null && !dmpEntity.getEntityDois().isEmpty())
                                text = text.replace("{ARGOS.DMP.DOI}", dmpEntity.getEntityDois().getFirst().getDoi());
                            else
                                text = text.replace("{ARGOS.DMP.DOI}", "-");
                            r.setText(text, 0);
                        }
                        if(text.contains("{ARGOS.DMP.LAST_MODIFIED}")){
                            DateTimeFormatter formatter = DateTimeFormatter.ofPattern("dd/MM/yyyy").withZone(ZoneId.systemDefault());
                            text = text.replace("{ARGOS.DMP.LAST_MODIFIED}", formatter.format(dmpEntity.getUpdatedAt()));
                            r.setText(text, 0);
                        }
                    }
                }
            }
        });
    }
}
